program plwav_a3;
{$M 8192,0,655360 }
uses dos,crt,cmostime;


  { Written by Piotr Fabian, 1994-02-12 }
  { PIOTRCF@star.iinf.gliwice.edu.pl    }

  { PLWAV_A3 plays a wavefile through PC Speaker or            }
  { a simple 8-bit D/A converter at base8 (see below).         }

  { This version uses up to 640KB (- (DOS + program) length)   }
  { memory for waveform data, divided into 8KB pages.          }

  { Playing a file through a digital to analog converter (DAC) }
  { is very simple - the only thing to do is to send samples   }
  { to DAC port in equal periods of time. Fortunately each PC  }
  { has an internal circuit, namely 8253 or 8254 which is used }
  { to generate "clock interrupts", usually 18.2 times per     }
  { second. This freqency can be changed, as well as the       }
  { clock interrupt routine. So playing a wavefile can be      }
  { performed by setting the interrupt frequency equal to      }
  { sampling frequency of wavefile and and sending samples     }
  { to DAC every time interrupt is called.                     }
  { It's simple to do using a "true" DAC, which converts       }
  { a digital value (e.g. 0..255) to analog signal.            }
  { Unfortunately the PC Speaker is connected to one of        }
  { DIGITAL outputs of 8253 (8254) circuit (one 8253 circuit   }
  { has three channels individually programmed). This means,   }
  { that only two values are possible: "0" or "1" (about 0 or  }
  { 5 Volts). But the PC Speaker has his inertia and if the    }
  { frequency of square waveform ("square" because only "0"    }
  { and "1" are possible) is high enough, the diaphragm        }
  { achives an average position.                               }
  { A small example will clear this method (I hope so):        }

  { The waveform modulates the width of square impulses        }
  { generated in short timeperiods:                            }

  { ORIGINAL WAVEFORM

       **************
    ***              *********
****                          *******                                  ******
                                     *******                    *******
                                            *****        *******
                                                 ********

 PULSE WIDTH MODULATED SQUARE WAVE
  +-------+  +------+   +------+   +-----+    +-+        +--+       +-----+
  |       |  |      |   |      |   |     |    | |        |  |       |     |
  |       |  |      |   |      |   |     |    | |        |  |       |     |
--+       +--+      +---+      +---+     +----+ +--------+  +-------+     +--

--+----------+----------+----------+----------+----------+----------+------>
                                                                       time
                        <---------->
                         1/(sampling_frequency)

  { Such a square waveform can be generated using TWO channels }
  { of 8253: one channel generates interrupts, the second      }
  { generates a single high impulse, which length depends      }
  { on the value of current sample.                            }
  { E.g. if you have following samples: 16, 128, 200 (whole    }
  { range: 0..255) and sampling frequency = 10000 Hz, you can  }
  { switch the speaker to:                                     }
  { "1" for 0.0063 ms and "0" for 0.0937 ms,                   }
  { "1" for 0.0500 ms and "0" for 0.0500 ms,                   }
  { "1" for 0.0784 ms and "0" for 0.0216 ms.                   }
  { The 8253 has to be programmed every time new sample        }
  { is being processed (8000..30000 times per second!).        }
  { I use a special mode of 8253 - "generate a single impulse".}
  { The length of this impulse ("1") depends on the sample.    }
  { A special table (tab1) containes values to be send to 8253 }
  { for different lengths of impulses.                         }

  { Usage:                                                     }
  {   plwav_a3 <file.wav> {p|a}
  {                        | |                                 }
  {                        | +---- use DAC at base8            }
  {                        +------ use PC Speaker              }

const
  intfreq:word=9000;   { sampling frequency = interrupt frequency   }
  base=$300;           { base address of my AD/DA hardware card     }
  base8=base+8;        { base address of DAC port                   }
  ctcclk=14318180/12;  { clock frequency for 8253 in Hz             }
  inint:boolean=false;
  tooslow:boolean=false;
  wzm:integer=3;       { amplification; should be equal 1 for       }
                       { D/A converter and e.g. 3 for PC speaker    }

type
  ttab=array[0..8191] of byte;
  ptab=^ttab;

var
  akt:byte;
  taa:array[0..79] of ptab;  { pointers to 8KB blocks               }
  max_t:integer;  { index of last used element in taa[]             }
  l_used:integer; { number of valid values in w taa[max_t]^         }
  max:word;
  w,gl_w:word;
  old8:pointer;   { previous interrupts vector                      }
  I,Nsamp:longint;
  K_ta,K_offs:word;
  p_seg,p_ofs:word;
  p,q,intp:integer;
  ab,b:byte;
  tab1:array[0..255] of byte;
  tab2:array[0..255] of integer;
  podzial:Word;
  sampfreq:longint;
  naz   : string[255];
  intrm : byte;
  outdev: integer;

{ readh(var f:FILE;var sampfreq,len:longint):boolean                  }
{                                                                     }
{ Reads WAVE header, returns true if no errors occured.               }
{ typ = 2*bits_per_sample+(numb_of_channels-1)                        }

function readh(var f:FILE;var sampfreq,len:longint; var typ:word):boolean;
var i,rlen,flen,dlen,nast,DWbuf:longint;
    BytesPerSamp,Wbuf,k:word;
    valid:boolean;
    c:string[16];

begin
  readh:=false;
  c[0]:=#4;

  blockread(f,c[1],4{bytes},k);      { 'RIFF'           }
  valid:= c='RIFF';
  if not(valid) or (k<>4) then exit;

  blockread(f,rlen,4,k);             {  rlen            }
  if k<>4 then exit;

  blockread(f,c[1],4{bytes},k);      { 'WAVE'           }
  valid:= c='WAVE';
  if not(valid) or (k<>4) then exit;

  blockread(f,c[1],4{bytes});        { 'fmt '           }
  valid:= c='fmt ';
  if not(valid) or (k<>4) then exit;

  blockread(f,flen,4{bytes},k);      {  flen            }
  if k<>4 then exit;

  nast:=FilePos(f)+flen;

  blockread(f,Wbuf,2{bytes},k);      {  wFormatTag      }
  if k<>2 then exit;
  if Wbuf<>1 then { format inny ni WAVE_FORMAT_PCM }
    exit;

  blockread(f,Wbuf,2{bytes},k);      {  wChannels       }
  if k<>2 then exit;
  case Wbuf of
    1: typ:=0;  { mono   }
    2: typ:=1;  { stereo }
    else exit;  { bd!  }
  end;

  blockread(f,DWbuf,4{bytes},k);    {  dwSamplesPerSec  }
  if k<>4 then exit;
  sampfreq:=DWbuf;
  if (sampfreq<4670) or (sampfreq>44100) then
    begin
      WriteLn('Sampling frequency not supported: ',
               sampfreq);
      Writeln('Supported frequencies are: 4670 Hz .. 44100 Hz');
      exit;
    end;

  blockread(f,DWbuf,4{bytes},k);    {  AvgBytesPerSec   }
  if k<>4 then exit;
  blockread(f, Wbuf,2{bytes},k);    {  wBlockAlign      }
  if k<>2 then exit;

  blockread(f, Wbuf,2{bytes},k);    {  wBitsPerSample   }
  if k<>2 then exit;
  if (Wbuf<1) or (Wbuf>16) then exit;
  typ:=typ+2*Wbuf;

  { szukaj 'data' }
  seek(f,nast);

  blockread(f,c[1],4{bytes},k);     { 'data'            }
  valid:= c='data';
  if not(valid) or (k<>4) then exit;

  blockread(f,dlen,4{bytes},k);     {  AvgBytesPerSec   }
  if k<>4 then exit;

  if typ div 2 < 9 then BytesPerSamp:=1 else BytesPerSamp:=2;
  if typ and 1 = 1 then BytesPerSamp:=2*BytesPerSamp;
  i:=FileSize(f)-FilePos(f);

  if (i <> dlen) then
    writeln('Number of samples declared in the header differs from the real number.');

  len:=i div BytesPerSamp;

  readh:=true

end;

procedure enableirq0; { turn IRQ 0 on }
var pom,maska:byte;
begin
  maska:=254;
  pom:=port[$21];     { 8259                     }
  pom:=pom and maska; { clear (enable) LSB       }
  port[$21]:=pom;     
end;

procedure disableirq0; { turn IRQ 0 off }
var pom,maska:byte;
begin
  maska:=1;
  pom:=port[$21];     { 8259                     }
  pom:=pom or maska;  { set LSB (disable)        }
  port[$21]:=pom;     
end;

procedure di;  { turn all interrupts off, except IRQ 0 }
begin
  intrm:=port[$21];
  port[$21]:=255- {3} 1;
end;

procedure ei;
begin
  port[$21]:=intrm;
end;

procedure tim_init;   { set counter mode }
begin
  port[$43]:=$34
end;


procedure init_freq(f:word); { set frequency }
begin
  if f>0 then gl_w:=round(ctcclk/f) else gl_w:=0;
  port[$40]:=Lo(gl_w);
  port[$40]:=Hi(gl_w);
end;


procedure gl_on;  { turn PC Speaker on }
begin
  port[$61]:=3 or port[$61];
end;

procedure gl_off; { turn PC Speaker off }
begin
  port[$61]:=$fc and port[$61];
end;


{ readsam8                                          }
{                                                   }
{ returns 8-bit sample; p=false if an error occured }

function readsamp8(var f:FILE; typ:word; var p:boolean):byte;
var chL,chR : longint;
    Ibuf    : integer;
    Bbuf    : byte;
    k       : word;
begin
  { Most often: mono, 8 bits }
  if typ=16 then
    begin
      blockread(f,Bbuf,1,k);
      if k<>1 then begin p:=false; readsamp8:=128; exit end;
      p:=true;
      readsamp8:=Bbuf;
      exit;
    end;
  if (typ and 1)=0 then { mono }
    begin
      if typ div 2 < 9 then { one byte for sample }
        begin
          blockread(f,Bbuf,1,k);
          if k<>1 then begin p:=false; readsamp8:=128; exit end;
          p:=true;
          readsamp8:=Bbuf;
          exit;
        end
      else                  { two bytes per sample  }
        begin
          blockread(f,Ibuf,2,k);
          if k<>2 then begin p:=false; readsamp8:=128; exit end;
          p:=true;
          readsamp8:=128 + Ibuf div 256;
          exit;
        end
    end
  else  { stereo }
    begin
      if typ div 2 < 8 then { one byte per sample }
        begin
          blockread(f,Bbuf,1,k);
          if k<>1 then begin p:=false; readsamp8:=128; exit end;
          Ibuf:=Bbuf;
          blockread(f,Bbuf,1,k);
          if k<>1 then begin p:=false; readsamp8:=128; exit end;
          Ibuf:=Ibuf+Bbuf;
          p:=true;
          readsamp8:=Ibuf div 2;
          exit;
        end
      else                  { two bytes per sample }
        begin
          blockread(f,Ibuf,2,k);
          if k<>2 then begin p:=false; readsamp8:=128; exit end;
          chL:=Ibuf;
          blockread(f,Ibuf,2,k);
          if k<>2 then begin p:=false; readsamp8:=128; exit end;
          chR:=Ibuf;
          p:=true;
          readsamp8:=128 + ( ( chL + chR ) div 512);
          exit;
        end
    end

end;


procedure fillt1;
var
  f         : file;
  l,zostalo : longint;
  i,j,
  odczytano : integer;
  typ       : word;          { type of WAVE file }
  ok        : boolean;
begin
  if ParamCount>=1 then naz:=ParamStr(1) else naz:='dane.wav';
  if Pos('.',naz)=0 then naz:=naz+'.wav';
{$I-}
  assign(f,naz);
  reset(f,1);
  if IOResult<>0 then
    begin
      Write('Error opening input file: ',naz);
      ClrEol;
      halt(1)
    end;
{$I+}
  if not(readh(f,sampfreq,l,typ)) then
    begin writeln('Bad format of input file (.wav) .'); Halt(1) end;
  intfreq:=sampfreq;
  zostalo:=l;
  max_t:=-1;
  while zostalo>0 do
  begin
    inc(max_t);
    if MaxAvail<8200 then
      begin
        for j:=0 to max_t-1 do dispose(taa[j]);
        WriteLn('Out of memory (wave to long?)...');
        Halt(1);
      end;
    new(taa[max_t]);
    if typ=16 then  { mono, 8 bits }
      begin
        blockread(f,taa[max_t]^,8192,odczytano);
      end
    else
      begin
        i:=0;
        repeat
          taa[max_t]^[i]:=readsamp8(f,typ,ok);
          if ok then inc(i);
        until not(ok) or (i=8191);
        odczytano:=i+1;
      end;

    zostalo:=zostalo-odczytano;
    if (zostalo>0) and (odczytano=0) then zostalo:=0;

  end;

  WriteLn('Memory left: ', MaxAvail,' bytes          ');

  close(f);
  l_used:=odczytano;
  max:=8191;  { only for K_offs=0..max_t-1; then max=odczytano-1 }
  K_offs:=max;
  K_ta:=word(-1);
  Nsamp:=longint(8192)*max_t-1+l_used;

end;

{$F+}

procedure obs_passPC; { INT 8h (clock interrupt IRQ 0) }
assembler;
   asm
    push ax
    push ds
    push es
    push bx
    mov ax,seg(akt)
    mov ds,ax
    cmp akt,0
    je @nicnierob
    mov bx,K_offs
    inc bx
    cmp bx,max
    jna @d

    { One block (8KB) processed!}
    mov bx,K_ta
    inc bx
    cmp bx,max_t
    jne @d4 { jump if not the last block or max_t+1 }
    mov ax,l_used
    dec ax
    mov max,ax
    stc          { don't jump to @d5                             }
  @d4:
    ja @d5      { jump if K_ta>max_t                             }
    mov K_ta,bx { update the index of block                      }
    shl bx,1    { one far pointer takes...                       }
    shl bx,1    { ...four bytes                                  }
    add bx,offset(taa)
    mov ax,[bx]
    mov p_ofs,ax
    inc bx
    inc bx
    mov ax,[bx]
    mov p_seg,ax
    mov bx,0    { mov K_offs,0 - clear index inside a block      }
    jmp @d
  @d5:
    mov akt,0
    jmp @nicnierob


  @d:
    mov K_offs,bx
    mov ax,p_seg
    mov es,ax
    add bx,p_ofs
    mov al,es:[bx]
    mov bx,offset(tab1)
    xlat
    mov bl,al
    mov al,128+32+16
    out $43,al
    mov al,bl
    out $42,al
    xor al,al
    out $42,al

   @nicnierob:
    mov al,$20
    out $20,al
    pop bx
    pop es
    pop ds
    pop ax
    iret
end;

procedure obs_passADDA; { INT 8h (clock interrupt IRQ 0) }
assembler;
   asm
    push ax
    push ds
    push es
    push bx
    push dx
    mov ax,seg(akt)
    mov ds,ax
    cmp akt,0
    je @nicnierob
    mov bx,K_offs
    inc bx
    cmp bx,max
    jna @d

    { One block (8KB) processed!!!}
    mov bx,K_ta
    inc bx
    cmp bx,max_t
    jne @d4 { jump if not the last block or max_t+1 }
    mov ax,l_used
    dec ax
    mov max,ax
    stc          { do not jump to @d5                 }
  @d4:
    ja @d5      { jump if K_ta>max_t                             }
    mov K_ta,bx { update the index of used block                 }
    shl bx,1    { one far pointer occupies...                    }
    shl bx,1    { ...four bytes                                  }
    add bx,offset(taa)
    mov ax,[bx]
    mov p_ofs,ax
    inc bx
    inc bx
    mov ax,[bx]
    mov p_seg,ax
    mov bx,0    { mov K_offs,0 - clear index inside a block }
    jmp @d
  @d5:
    mov akt,0
    jmp @nicnierob


  @d:
    mov K_offs,bx
    mov ax,p_seg
    mov es,ax
    add bx,p_ofs
    mov al,es:[bx]

    mov dx,base8
    out dx,al

   @nicnierob:
    mov al,$20
    out $20,al
    pop dx
    pop bx
    pop es
    pop ds
    pop ax
    iret
end;


procedure init_tab2;
var i,j:byte;k,d:integer;
begin
  for i:=0 to 255 do
    begin
      j:=i;
      { increase volume }
      k:=j;
      k:=round(128+wzm*(k-128));
      if k<0 then k:=0;
      if k>255 then k:=255;
      j:=k;
      d:=round(gl_w*j/255);
      if d=0 then inc(d);
      if d>255 then
        begin
	  WriteLn('Error in init_tab2');
	  exit
	end;
      tab1[i]:=d;
    end
end;


procedure info;
begin
  WriteLn;
  WriteLn;
  WriteLn('This program plays WAVE audio files through PC speaker');
  WriteLn('or a simple 8-bit D/A converter (at $308).');
  WriteLn;
  WriteLn('Usage:');
  WriteLn('');
  WriteLn('plwav_a3 <file.wav> {p|a}');
  WriteLn('                     | |');
  WriteLn('                     | +- use D/A converter');
  WriteLn('                     +--- use PC speaker (default)');
  WriteLn;
  WriteLn('The speed of your computer is not checked - if too slow, the program');
  WriteLn('may not work properly. Also doesn''t work as DOS application in Windows.');
  WriteLn;
  WriteLn('Piotr Fabian, P.O.Box 12, 44-102 Gliwice 2, piotrcf@star.iinf.gliwice.edu.pl');
  WriteLn(#10#10);
end;


{ graj                                              }
{                                                   }
{ Play once the waveform from memory                }
{                                                   }

procedure graj;
begin
  ClrEol;
  {write('Playing! Press ESC to break, other key to rewind...'#13);}
  write('Playing! Please wait, the keyboard is locked...'#13);
  max:=8191;  { only for K_offs=0..max_t-1; then max=odczytano-1 }
  K_offs:=max;
  K_ta:=word(-1);

  akt:=1;  { turn activity flag on }
  if outdev=1 then gl_on;
  getintvec(8,old8);
  disableirq0;
  port[$43]:=128+32+16+0; { last (third) counter, count mode             }
  port[$42]:=0;           { clear...                                     }
  port[$42]:=0;           { ...counter                                   }

  port[$43]:=128+32+0;    { last (third) counter, count mode             }
  port[$42]:=0;

  case outdev of
    1: setintvec(8,@obs_passPC);
    2: setintvec(8,@obs_passADDA);
  end;
  tim_init;
  init_freq(intfreq);     { set interrupts frequency                     }
  init_tab2;
  enableirq0;
  di;

  asm
  @a1:
   { cli }
    cmp akt,0
    je @kon
    mov ax,0
    mov es,ax
    mov ah,1                   { function 1 }
    { pushf }
    { call [dword ptr es:88] } { ...int 16h - read the keaboard  }
    { jnz @kon }               { key pressed => break }
    jmp @a1
    @kon:
   { sti }
  end;
  ei;
  disableirq0;
  tim_init;
  init_freq(0);
  setintvec(8,old8);
  enableirq0;
  if outdev=1 then gl_off;
  ClrEol;

end;

begin

  ClrScr;
  info;
  outdev:=1; { PC speaker }

  Write('Wait - reading samples...'#13);
  fillt1;
  if ParamCount>1 then
    begin
      if (ParamStr(2)='p') or (ParamStr(2)='P') then outdev:=1;
      if (ParamStr(2)='a') or (ParamStr(2)='A') then outdev:=2;
    end;
  while KeyPressed do if ReadKey=' 'then;
  Write('Press a key...');
  ClrEol;
  if outdev=2 then wzm:=1 else wzm:=3;
  repeat until KeyPressed;
      if ReadKey=' ' then;
  ClrScr;

  WriteLn('File:                                      ',naz);
  WriteLn;

  WriteLn('Interrupts (8h) frequency:         about   ',intfreq:5,' Hz');

  I:=0;  ab:=128;
  if intfreq>0 then gl_w:=round(ctcclk/intfreq) else gl_w:=0;
  init_tab2;

  WriteLn('Sampling frequency:                        ',
           round(ctcclk/gl_w):5,' Hz');
  WriteLn('CTCCLK ticks between interrupts:           ',gl_w:5);
  WriteLn('Dynamic (with PC speaker):                 ',
           20*ln(gl_w)/ln(10):5:1,' dB');
  WriteLn('Number of samples:                         ',
           Nsamp);
  WriteLn('Length of wave:                            ',
           Nsamp/intfreq:6:2,' s');
  Write(#10#10);

  while KeyPressed do ReadKey;

  repeat
    graj;
    while KeyPressed do ReadKey;
    ClrEol;
    write(
     'Press ESC to exit or any other key to play'#13);
    repeat until KeyPressed;
  until ReadKey=#27;

  for i:=0 to max_t do dispose(taa[i]);

  ClrEol;
  writeln('uff...');

  while KeyPressed do if ReadKey=' 'then;
  nosound;
  copycmostime;

end.
